#include <htc.h>

// Hardware support.
#include "../hardware/display.h"
#include "../hardware/buttons.h"
#include "../hardware/resistors.h"
#include "../utility/delay.h"

#define HARD_CODED_DELAYS 0

enum gate_delays {
	GATE_DELAY_1MS,
	GATE_DELAY_5MS,
	GATE_DELAY_10MS,
	GATE_DELAY_50MS,
	GATE_DELAY_100MS,
	GATE_DELAY_1S,
};

void tmr2_set_postscale(unsigned char postscale) {
	--postscale;
	TOUTPS0 = postscale;
	postscale >>= 1;
	TOUTPS1 = postscale;
	postscale >>= 1;
	TOUTPS2 = postscale;
	postscale >>= 1;	
	TOUTPS3 = postscale;
}

void frequency(void) {

	enum gate_delays gate_delay = GATE_DELAY_1MS;

	const unsigned int gate_delay_ticks[] = {
		   2, // 1ms
		  10, // 5ms
		  20, // 10ms
		 100, // 50ms
		 200, // 100ms
		2000, // 1s
	};

	// Use TMR2 to time the gate.
	TMR2ON = 1;
	// Use a prescaler of 1:4.
	// This results in a timer frequency of (Fosc/4)/4 = (20MHz/4)/4 = 1.25MHz.
	T2CKPS1 = 0; T2CKPS0 = 1;
	// Set the period of the counter to 100us.
	PR2 = 124; // (1.25MHz * 100us) - 1
	// Set the postscaler to 5 to have an interrupt period of 500us.
	tmr2_set_postscale(5);

	// Run the frequency meter until both buttons are pressed.
	while (BUTTON_1 || BUTTON_2) {

		// Gate the timer input.
		TRIS_CLAMP = 0;
		CLAMP = 0;		

		// Calculate the length of the delay in 500us ticks.
		unsigned int delay_ticks = gate_delay_ticks[gate_delay];

		// Don't let interrupts interfere with the timing.
		display_off_synchronised();
		di();

		// Use an external clock source.
		T0CS = 1;
	
		// Assign the prescaler to the timer.
		PSA = 0;
		// Prescaler = 1:256.
		PS2 = PS1 = PS0 = 1;
	
		// Clear the timer and prescaler.
		TMR0 = 0;

#if HARD_CODED_DELAYS
		
		display_off();
		
		switch (gate_delay) {
			case GATE_DELAY_1MS:
				TRIS_CLAMP = 1;
				__delay_ms(1);
				TRIS_CLAMP = 0;
				break;
			case GATE_DELAY_5MS:
				TRIS_CLAMP = 1;
				__delay_ms(5);
				TRIS_CLAMP = 0;
				break;
			case GATE_DELAY_10MS:
				TRIS_CLAMP = 1;
				__delay_ms(10);
				TRIS_CLAMP = 0;
				break;
			case GATE_DELAY_50MS:
				TRIS_CLAMP = 1;
				__delay_ms(25);
				__delay_ms(25);
				TRIS_CLAMP = 0;
				break;
			case GATE_DELAY_100MS:
				TRIS_CLAMP = 1;
				__delay_ms(25);
				__delay_ms(25);
				__delay_ms(25);
				__delay_ms(25);
				TRIS_CLAMP = 0;
				break;
		}

	
#else

		// Align to TMR2 before starting.
		TMR2 = 0;
		TMR2IF = 0;
		while (!TMR2IF);
		// Open the input gate.
		TRIS_CLAMP = 1;
		// The alignment counts as the first tick.	
		--delay_ticks;
		
		for (;;) {
			TMR2IF = 0;
			while (!TMR2IF);
			--delay_ticks;
			if (delay_ticks == 0) {
				// Wait for the final tick.
				TMR2IF = 0;
				while (!TMR2IF);
				// Close the input gate.
				TRIS_CLAMP = 0;
				break;
			} else {
				// Refresh the display in the background.
				display_output_segment();
			}
		}

#endif
	
		// Did we overflow?
		unsigned char overflowed = T0IF;

		// Calculate the prescaler.
		unsigned char previous_tmr0 = TMR0;
		unsigned char prescaler = 1;
		do {
			display_output_segment();
			--prescaler;
			CLAMP = 1;
			CLAMP = 0;
		} while (TMR0 == previous_tmr0);
	
		unsigned int ticks = (previous_tmr0 << 8) | prescaler;

		// Assign the prescaler to the WDT.
		PSA = 0;
		// Prescaler = 1:1.
		PS2 = PS1 = PS0 = 0;

		// Switch back to interrupt-driven display.
		T0CS = 0;
		TMR0 = 0;
		TMR0IF = 0;
		ei();
		display_on();

#if HARD_CODED_DELAYS

		display_on();

#endif

		// Release the timer input.
		TRIS_CLAMP = 1;

		// Calculate the frequency.
		unsigned long frequency = (ticks * 2000UL) / gate_delay_ticks[gate_delay];

		if (!overflowed) {
			for (unsigned char delay = 0; delay < 75 && (BUTTON_1 || BUTTON_2); ++delay) {
				/*if (!BUTTON_1) {
					display_print_decimal(frequency / 10000);
					display_data[0] |= 0x80;
					display_data[1] |= 0x80;
					// Wait for button 1 to be released.
					delay_ms(10);
					while (!BUTTON_1);
					// Wait for button 1 to be pressed again.
					delay_ms(10);
					while (BUTTON_1);
					display_print_decimal(frequency % 10000);
					display_data[2] |= 0x80;
					display_data[3] |= 0x80;
					// Wait for button 1 to be released.
					delay_ms(10);
					while (!BUTTON_1);
					// Wait for button 1 to be pressed again.
					delay_ms(10);
					while (BUTTON_1);
					// Wait for button 1 to be released.
					delay_ms(10);
					while (!BUTTON_1);
					delay_ms(10);
				} else {*/
					display_print_float(frequency);
				//}
				delay_ms(10);
			}
		} else {
			display_clear();
		}

		// Adjust the gate delay according to the current frequency.
		if (overflowed) {
			gate_delay = GATE_DELAY_1MS;
		} else if (frequency < 50e3) {
			gate_delay = GATE_DELAY_1S;
		} else if (frequency < 500e3) {
			gate_delay = GATE_DELAY_100MS;
		} else if (frequency < 1e6) {
			gate_delay = GATE_DELAY_50MS;
		} else if (frequency < 5e6) {
			gate_delay = GATE_DELAY_10MS;
		} else if (frequency < 10e6) {
			gate_delay = GATE_DELAY_5MS;
		} else {
			gate_delay = GATE_DELAY_1MS;
		}

	}

	// We don't need TMR2 any more.
	TMR2ON = 0;
}