#include <htc.h>
#include "../hardware/display.h"
#include "../hardware/buttons.h"
#include "../utility/one-wire.h"
#include "../utility/delay.h"

unsigned char try_get_temperature(float* temperature) {

	// Store data/CRC for long transfers.
	unsigned char data[8];
	unsigned char crc;

	// Get the ROM ID.
	if (one_wire_reset()) return 0;
	one_wire_write_byte(ONE_WIRE_COMMAND_READ_ROM);
	for (unsigned char p = 0; p < 7; ++p) {
		data[p] = one_wire_read_byte();
	}
	crc = one_wire_read_byte();

	// Check the CRC
	if (one_wire_calculate_crc(data, 7) != crc) return 0;

	// Check the device family code (expected: 0x28 for DS18B20).
	if (data[0] != 0x28) return 0;

	// Write configuration data to the thermometer.
	if (one_wire_reset()) return 0;
	one_wire_write_byte(ONE_WIRE_COMMAND_SKIP_ROM);
	one_wire_write_byte(0x4E); // "Write scratchpad".
	one_wire_write_byte(0x00); // TH
	one_wire_write_byte(0x00); // TL
	one_wire_write_byte(0x7F); // Configuration (R1:R0 = 0b11 for a 12-bit resolution).
	
	// Start temperature conversion.
	if (one_wire_reset()) return 0;

	one_wire_write_byte(ONE_WIRE_COMMAND_SKIP_ROM);
	one_wire_write_byte_powered(0x44); // "Convert T".

	// Delay for at least 750ms
	for (unsigned char d = 0; d < 80; ++d) {
		if (!(BUTTON_1 || BUTTON_2)) {
			one_wire_init();
			return 0;
		}
		delay_ms(10);
	}

	// Disconnect the power by reinitialising the 1-Wire bus.
	one_wire_init();

	// Read the scratchpad.
	if (one_wire_reset()) return 0;
	one_wire_write_byte(ONE_WIRE_COMMAND_SKIP_ROM);
	one_wire_write_byte(0xBE); // "Read scratchpad".
	delay_ms(10);

	// Read all eight scratchpad bytes and its CRC.
	for (unsigned char p = 0; p < 8; ++p) {
		data[p] = one_wire_read_byte();
	}
	crc = one_wire_read_byte();

	if (one_wire_calculate_crc(data, 8) != crc) return 0;

	// Convert the temperature to a floating point value and return successfully.
	*temperature = (int)((data[1] << 8) | data[0]) * 0.0625f;
	return 1;	
}

void temperature(void) {

	// We'll be using the 1-Wire bus.
	one_wire_init();

	while (BUTTON_1 || BUTTON_2) {

		// Take the temperature.
		float temperature;
		unsigned char got_temperature = try_get_temperature(&temperature);
		if (got_temperature) {
			// Display the temperature.
			display_print_float(temperature);	
		} else {
			// Display '----'.
			display_set_display(0b01000000, 0b01000000, 0b01000000, 0b01000000);
		}
	}
}